package meerkat.bulletinboard.sqlserver;

import java.io.File;
import java.util.Arrays;
import java.util.List;

import com.google.protobuf.ProtocolStringList;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import meerkat.bulletinboard.BulletinBoardServer;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto.Signature;
import meerkat.protobuf.Crypto.SignatureVerificationKey;
import meerkat.crypto.Digest;
import meerkat.crypto.concrete.SHA256Digest;

public abstract class BulletinBoardSQLServer implements BulletinBoardServer{
	
	protected Connection connection;
	
	protected Digest digest;
	
	protected List<SignatureVerificationKey> trusteeSignatureVerificationArray;
	protected int minTrusteeSignatures;

	protected List<List<SignatureVerificationKey>> pollingCommitteeSignatureVerificationKeyArray;
	protected int minCommiteeSignatures;
	
	/**
	 * This method initializes the signatures but does not implement the DB connection.
	 * Any full (non-abstract) extension of this class should
	 * 1. Establish a DB connection, and
	 * 2. Call this procedure
	 */
	@Override
	public void init(File meerkatDB) throws CommunicationException {
		// TODO write signature reading part.
		
		digest = new SHA256Digest();
	}

	/**
	 * This method verifies the authenticity of the received message based on
	 * the stored signatures.
	 * 
	 * @param msg
	 *            is the message to authenticate (containing the signature).
	 * @return TRUE if the message is authenticated and FALSE otherwise.
	 */
	private boolean verifyMessage(BulletinBoardMessage msg) {
		//TODO: Replace with actual verification.
		return true;
	}
	
	/**
	 * This procedure makes sure that all tags in the given list have an entry in the tags table. 
	 * @param tags
	 */
	protected abstract void insertNewTags(String[] tags) throws SQLException;
	
	/**
	 * This procedure is used to convert a boolean to a BoolMsg.
	 * @param b is the boolean to convert.
	 * @return a ProtoBuf message with boolean payload.
	 */
	private BoolMsg boolToBoolMsg(boolean b){
		return BoolMsg.newBuilder()
			.setValue(b)
			.build();
	}
	
	@Override
	public BoolMsg postMessage(BulletinBoardMessage msg) throws CommunicationException {
		
		if (!verifyMessage(msg)) {
			return boolToBoolMsg(false);
		}
		
		PreparedStatement pstmt;
		ResultSet rs;
		String sql;
		byte[] msgID;
		long entryNum;
		
		ProtocolStringList tagList;
		String[] tags;
		
		List<Signature> signatureList;
		Signature[] signatures;
		
		// Calculate message ID (depending only on the the unsigned message)
		
		digest.reset();
		digest.update(msg.getMsg());
		
		msgID = digest.digest();
		
		// Add message to table if needed and store entry number of message.
		
		try {
			
			sql = "SELECT EntryNum From MsgTable WHERE MsgId = ?";
			pstmt = connection.prepareStatement(sql);
			pstmt.setBytes(1, msgID);
			rs = pstmt.executeQuery();
			
			if (rs.next()){
				
				entryNum = rs.getLong(1);
				
			} else{
				
				sql = "INSERT INTO MsgTable (MsgId, Msg) VALUES(?,?)";
			    pstmt = connection.prepareStatement(sql);
			    pstmt.setBytes(1, msgID);
			    pstmt.setBytes(2, msg.getMsg().toByteArray());
			    pstmt.executeUpdate();
			    
			    rs = pstmt.getGeneratedKeys();
			    rs.next();
			    entryNum = rs.getLong(1);
				
			}
		    
		    pstmt.close();
		    
		} catch (SQLException e) {
			throw new CommunicationException("Error inserting into MsgTable: " + e.getMessage());
		}
		
		// Retrieve tags and store new ones in tag table.
		
		try {
			
		    tagList = msg.getMsg().getTagList();
		    tags = new String[tagList.size()];
		    tags = tagList.toArray(tags);
		    
		    insertNewTags(tags);
		    
		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}
		
		// Connect message to tags.
		
		try{
		    sql = "INSERT OR IGNORE INTO MsgTagTable (TagId, EntryNum) SELECT TagTable.TagId, ? AS EntryNum FROM TagTable WHERE Tag = ?";
	    	pstmt = connection.prepareStatement(sql);
		    
	    	pstmt.setLong(1, entryNum);
	    	
		    for (String tag : tags){
		    	pstmt.setString(2, tag);
		    	pstmt.addBatch();
		    }
		    
		    pstmt.executeBatch();
		    pstmt.close();
			
		} catch (SQLException e) {
			throw new CommunicationException("Error Linking tags: " + e.getMessage());
		}
		
		// Retrieve signatures.
		
	    signatureList = msg.getSigList();
	    signatures = new Signature[signatureList.size()];
	    signatures = signatureList.toArray(signatures);
		
		// Connect message to signatures.
		
		try{
		    sql = "INSERT OR IGNORE INTO SignatureTable (EntryNum, SignerId, Signature) VALUES (?,?,?)";
	    	pstmt = connection.prepareStatement(sql);
		    
	    	pstmt.setLong(1, entryNum);
	    	
		    for (Signature sig : signatures){
		    	
		    	pstmt.setBytes(2, sig.getSignerId().toByteArray());
		    	pstmt.setBytes(3, sig.toByteArray());
		    	pstmt.addBatch();
		    }
		    
		    pstmt.executeBatch();
		    pstmt.close();
			
		} catch (SQLException e) {
			throw new CommunicationException("Error Linking tags: " + e.getMessage());
		}

		return boolToBoolMsg(true);
	}
	
	public String testPrint(){
		
		String s = "";
		
		try {
	    
		    Statement statement = connection.createStatement();
			ResultSet rs = statement.executeQuery("select * from MsgTable");
			while (rs.next()) {
				// read the result set
				s += "entry = " + rs.getInt("EntryNum") + " \n";
				s += "id = " + Arrays.toString(rs.getBytes("MsgId")) + " \n";
				s += "msg = " +  Arrays.toString(rs.getBytes("Msg")) + " \n";
				s += "signer ID = " +  Arrays.toString(rs.getBytes("SignerId")) + "\t\n<BR>";
			}
			
			rs = statement.executeQuery("select * from TagTable");
			while (rs.next()) {
				// read the result set
				s += "Tag = " + rs.getString("Tag") + " \n";
				s += "TagId = " + rs.getInt("TagId") + "\t\n<BR>";
			}
			
			rs = statement.executeQuery("select * from MsgTagTable");
			while (rs.next()) {
				// read the result set
				s += "MsgId = " +  Arrays.toString(rs.getBytes("MsgId")) + " \n";
				s += "TagId = " + rs.getInt("TagId") + "\t\n<BR>";
			}
		} catch(SQLException e){
			s += "Error reading from DB";
		}
		
		return s;
	}

}
