package meerkat.bulletinboard.sqlserver;

import java.io.File;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;

import com.google.protobuf.InvalidProtocolBufferException;

import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto.Signature;
import meerkat.bulletinboard.sqlserver.BulletinBoardSQLServer;
import meerkat.comm.CommunicationException;

public class SQLiteBulletinBoardServer extends BulletinBoardSQLServer {

	protected static final int TIMEOUT = 20;
	
	/**
	 * This procedure initializes:
	 * 1. The database connection
	 * 2. The database tables (if they do not yet exist).
	 */
	
	private void createSchema() throws SQLException {
		Statement statement = connection.createStatement();
		statement.setQueryTimeout(TIMEOUT);
		
		statement.executeUpdate("CREATE TABLE IF NOT EXISTS MsgTable (EntryNum INTEGER PRIMARY KEY, MsgId BLOB UNIQUE, Msg BLOB)");
		
		statement.executeUpdate("CREATE TABLE IF NOT EXISTS TagTable (TagId INTEGER PRIMARY KEY, Tag varchar(50) UNIQUE)");
		statement.executeUpdate("CREATE TABLE IF NOT EXISTS MsgTagTable (EntryNum BLOB, TagId INTEGER, FOREIGN KEY (EntryNum) REFERENCES MsgTable(EntryNum), FOREIGN KEY (TagId) REFERENCES TagTable(TagId), UNIQUE (EntryNum, TagID))");
		
		statement.executeUpdate("CREATE TABLE IF NOT EXISTS SignatureTable (EntryNum BLOB, SignerId BLOB, Signature BLOB UNIQUE, FOREIGN KEY (EntryNum) REFERENCES MsgTable(EntryNum))");
		statement.executeUpdate("CREATE INDEX IF NOT EXISTS SignerIndex ON SignatureTable(SignerId)");
		
		statement.close();
	}
	
	@Override
	public void init(File meerkatDB) throws CommunicationException {
		
		try{
			String dbString = "jdbc:sqlite:" + meerkatDB.getPath();
			connection = DriverManager.getConnection(dbString);
			createSchema();
			
			super.init(meerkatDB);
				
		} catch (SQLException e) {
			
			throw new CommunicationException("Couldn't form a connection with the database" + e.getMessage());
			
		}
				
	}
	
	public void close() throws CommunicationException{
		
		try{
			connection.close();
		} catch (SQLException e) {
			
			throw new CommunicationException("Couldn't close connection to the database");
			
		}
		
	}

	@Override
	protected void insertNewTags(String[] tags) throws SQLException {
		
		PreparedStatement pstmt;
		String sql;
		
		try {
			
		    sql = "INSERT OR IGNORE INTO TagTable(Tag) VALUES (?)";
	    	pstmt = connection.prepareStatement(sql);
		    
		    for (String tag : tags){
		    	pstmt.setString(1, tag);
		    	pstmt.addBatch();
		    }
		    
		    pstmt.executeBatch();
		    pstmt.close();
		    
		} catch (SQLException e){
			throw new SQLException("Error adding new tags to table: " + e.getMessage());
		}
		
	}

	@Override
	public BoolMsg postMessage(BulletinBoardMessage msg) throws CommunicationException {
		return super.postMessage(msg);
	}

	@Override
	public BulletinBoardMessageList readMessages(MessageFilterList filterList) throws CommunicationException{

		PreparedStatement pstmt;
		ResultSet messages, signatures;
		
		long entryNum;
		BulletinBoardMessageList.Builder resultListBuilder = BulletinBoardMessageList.newBuilder();
		BulletinBoardMessage.Builder messageBuilder;
		
		String sql;
		String sqlSuffix = "";
		
		List<MessageFilter> filters = filterList.getFilterList();
		int i;
		
		boolean tagsRequired = false;
		boolean signaturesRequired = false;
		
		boolean isFirstFilter = true;
		
		// Check if Tag/Signature tables are required for filtering purposes.
		
		for (MessageFilter filter : filters){
			if (filter.getType() == FilterType.TAG){
				tagsRequired = true;
			} else if (filter.getType() == FilterType.SIGNER_ID){
				signaturesRequired = true;
			}
		}
		
		sql = "SELECT MsgTable.EntryNum, MsgTable.Msg FROM MsgTable";
		
		if (tagsRequired){
			sql += " INNER JOIN MsgTagTable ON MsgTable.EntryNum = MsgTagTable.EntryNum";
			sql += " INNER JOIN TagTable ON TagTable.TagId = MsgTagTable.TagId";
		}
		
		if (signaturesRequired){
			sql += " INNER JOIN SignatureTable ON SignatureTable.EntryNum = MsgTable.EntryNum";
		}
		
		// Add conditions.
		
		if (!filters.isEmpty()){
			sql += " WHERE";
			
			for (MessageFilter filter : filters){
				
				if (filter.getType().getNumber() != FilterType.MAX_MESSAGES_VALUE){
					if (isFirstFilter){
						isFirstFilter = false;
					} else{
						sql += " AND";
					}
				}
				
				switch (filter.getType().getNumber()){
					case FilterType.EXACT_ENTRY_VALUE:
						sql += " MsgTable.EntryNum = ?";
						break;
					case FilterType.MAX_ENTRY_VALUE:
						sql += " MsgTable.EntryNum <= ?";
						break;
					case FilterType.MAX_MESSAGES_VALUE:
						sqlSuffix += " LIMIT = ?";
						break;
					case FilterType.MSG_ID_VALUE:
						sql += " MsgTableMsgId = ?";
						break;
					case FilterType.SIGNER_ID_VALUE:
						sql += " SignatureTable.SignerId = ?";
						break;
					case FilterType.TAG_VALUE:
						sql += " TagTable.Tag = ?";
						break;
				}
			}
			
			sql += sqlSuffix;
		}
		
		// Make query.
		
		try {
	    	pstmt = connection.prepareStatement(sql);
		    
	    	// Specify values for filters.
	    	
	    	i = 1;
		    for (MessageFilter filter : filters){
		    	
		    	switch (filter.getType().getNumber()){
		    		
					case FilterType.EXACT_ENTRY_VALUE: // Go through.
					case FilterType.MAX_ENTRY_VALUE:
						pstmt.setLong(i, filter.getEntry());
						i++;
						break;
					
					case FilterType.MSG_ID_VALUE: // Go through.
					case FilterType.SIGNER_ID_VALUE:
						pstmt.setBytes(i, filter.getId().toByteArray());
						i++;
						break;
						
					case FilterType.TAG_VALUE:
						pstmt.setString(i, filter.getTag());
						break;
					
					// The max-messages condition is applied as a suffix. Therefore, it is treated differently.
					case FilterType.MAX_MESSAGES_VALUE:
						pstmt.setLong(filters.size(), filter.getMaxMessages());
						break;
					
		    	}
		    }
		    
		    // Run query.
		    
		    messages = pstmt.executeQuery();
		    
		    // Compile list of messages.
		    
		    sql = "SELECT Signature FROM SignatureTable WHERE EntryNum = ?";
		    pstmt = connection.prepareStatement(sql);
		    
		    while (messages.next()){
		    	
		    	// Get entry number and retrieve signatures.
		    	
		    	entryNum = messages.getLong(1);
		    	pstmt.setLong(1, entryNum);
		    	signatures = pstmt.executeQuery();
		    	
		    	// Create message and append signatures.
		    	
		    	messageBuilder = BulletinBoardMessage.newBuilder()
		    		.setEntryNum(entryNum)
		    		.setMsg(UnsignedBulletinBoardMessage.parseFrom(messages.getBytes(2)));
		    	
		    	while (signatures.next()){
		    		messageBuilder.addSig(Signature.parseFrom(signatures.getBytes(1)));
		    	}
		    	
		    	// Finalize message and add to message list.
		    	
		    	resultListBuilder.addMessage(messageBuilder.build());
		    	
		    }
		    
			pstmt.close();
		    
		} catch (SQLException e){
			throw new CommunicationException("Error reading messages from DB: " + e.getMessage());
		} catch (InvalidProtocolBufferException e) {
			throw new CommunicationException("Invalid data from DB: " + e.getMessage());
		}
		
		//Combine results and return.
		
		return resultListBuilder.build();
	}

}
