package meerkat.bulletinboard;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.List;
import java.util.Random;

import com.google.protobuf.ByteString;

import meerkat.comm.CommunicationException;
import meerkat.crypto.concrete.ECDSASignature;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.BulletinBoardAPI.FilterType;
import meerkat.protobuf.BulletinBoardAPI.MessageFilter;
import meerkat.protobuf.BulletinBoardAPI.MessageFilterList;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;

import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

public class GenericBulletinBoardServerTest {
	private BulletinBoardServer bulletinBoardServer;
	private ECDSASignature signers[];
	
	private Random random;
	
	private static String KEYFILE_EXAMPLE = "/certs/enduser-certs/user1-key-with-password-secret.p12";
	private static String KEYFILE_PASSWORD = "secret";
	
	public static String CERT1_PEM_EXAMPLE = "/certs/enduser-certs/user1.crt";
	
//	private static String KEYFILE_EXAMPLE2 = "/certs/enduser-certs/user2-key.pem";
	
	public void init(Class<?> cls) throws InstantiationException, IllegalAccessException, CertificateException, KeyStoreException, NoSuchAlgorithmException, IOException, UnrecoverableKeyException, CommunicationException{
		bulletinBoardServer = (BulletinBoardServer) cls.newInstance();

		bulletinBoardServer.init(File.createTempFile("meerkat-test", "db"));
		
		signers = new ECDSASignature[2];
        signers[0] = new ECDSASignature();
        signers[1] = new ECDSASignature();
		
		InputStream keyStream = getClass().getResourceAsStream(KEYFILE_EXAMPLE);
        char[] password = KEYFILE_PASSWORD.toCharArray();
        
        KeyStore.Builder keyStore = signers[0].getPKCS12KeyStoreBuilder(keyStream, password);
        signers[0].loadSigningCertificate(keyStore);
        
        signers[0].loadVerificationCertificates(getClass().getResourceAsStream(CERT1_PEM_EXAMPLE));
        
        random = new Random(0); // We use insecure randomness in tests for repeatability
	}
	
	private byte randomByte(){
		return (byte) random.nextInt();
	}
	
	private String randomString(){
		return new BigInteger(130, random).toString(32);
	}
	
	public void bulkTest() throws CommunicationException, SignatureException, InvalidKeyException, CertificateException, IOException{
		
		final int TAG_NUM = 5;					// Number of tags.
		final int MESSAGE_NUM = 32;				// Number of messages (2^TAG_NUM).
		final int BYTES_PER_MESSAGE_DATA = 50;	// Message size.
		
		String[] tags = new String[TAG_NUM];
		byte[][] data = new byte[MESSAGE_NUM][BYTES_PER_MESSAGE_DATA];
		
		UnsignedBulletinBoardMessage.Builder unsignedMsgBuilder;
		BulletinBoardMessage.Builder msgBuilder;
		
		int i,j;
		
		// Generate random data.
		
		for (i = 1 ; i <= MESSAGE_NUM ; i++){
			for (j = 0 ; j < BYTES_PER_MESSAGE_DATA ; j++){
				data[i-1][j] = randomByte();
			}
		}
		
		for (i = 0 ; i < TAG_NUM ; i++){
			tags[i] = randomString();
		}
		
		// Build messages.
		
		for (i = 1 ; i <= MESSAGE_NUM ; i++){
			unsignedMsgBuilder = UnsignedBulletinBoardMessage.newBuilder()
				.setData(ByteString.copyFrom(data[i-1]));
			
			// Add tags based on bit-representation of message number.
			
			int copyI = i;
			for (j = 0 ; j < TAG_NUM ; j++){
				if (copyI % 2 == 1){
					unsignedMsgBuilder.addTag(tags[j]);
				}
				
				copyI >>>= 1;
			}
			
			// Build message.
			
			msgBuilder = BulletinBoardMessage.newBuilder()
					.setMsg(unsignedMsgBuilder.build());
	        
			// Add signatures.
			
	        if (i % 2 == 1){
	        	signers[0].updateContent(msgBuilder.getMsg());
	        	msgBuilder.addSig(signers[0].sign());
	        }
			
	        // Post message.
	        
			bulletinBoardServer.postMessage(msgBuilder.build());
		}
		
		// Check tag mechanism
		
		for (i = 0 ; i < TAG_NUM ; i++){
			
			// Retrieve messages having tag i
			
			List<BulletinBoardMessage> messages = 
				bulletinBoardServer.readMessages(
					MessageFilterList.newBuilder()
						.addFilter(MessageFilter.newBuilder()
							.setType(FilterType.TAG)
							.setTag(tags[i])
							.build()
						)
						.build()
					)
					.getMessageList();
			
			// Assert that the number of retrieved messages is correct.
			
			assertThat(messages.size(), is(MESSAGE_NUM / 2));
			
			// Assert the identity of the messages.
			
			for (BulletinBoardMessage msg : messages){
				
				// Assert serial number and raw data.
				
				assertThat((msg.getEntryNum() >>> i) % 2 , is((long) 1));
				assertThat(msg.getMsg().getData().toByteArray(), is(data[(int) msg.getEntryNum() - 1]));
				
				// Assert signatures.
				
				if (msg.getEntryNum() % 2 == 1){
					signers[0].initVerify(msg.getSig(0));
			        signers[0].updateContent(msg.getMsg());
			        assertTrue("Signature did not verify!", signers[0].verify());
				}
			}
			
		}
		
	}
	
	public void close(){
		signers[0].clearSigningKey();
		signers[1].clearSigningKey();
	}
}
