package meerkat.crypto.concrete;

import com.google.protobuf.ByteString;
import com.google.protobuf.Message;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Crypto.Signature;
import org.bouncycastle.asn1.ASN1Integer;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.crypto.params.AsymmetricKeyParameter;
import org.bouncycastle.crypto.params.ECPrivateKeyParameters;
import org.bouncycastle.crypto.signers.DSAKCalculator;
import org.bouncycastle.crypto.signers.ECDSASigner;
import org.bouncycastle.crypto.signers.HMacDSAKCalculator;
import org.bouncycastle.jcajce.provider.asymmetric.util.ECUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.*;
import java.security.cert.*;
import java.security.interfaces.ECPrivateKey;


/**
 * Sign and verify digital signatures.
 *
 * Uses deterministic ECDSA signatures as per RFC 6979
 *
 * This class uses BouncyCastle directly, so will not work with arbitrary PKCS11 providers.
 *
 * This class is not thread-safe (each thread should have its own instance).
 */
public class ECDSADeterministicSignature extends ECDSASignature {
    final Logger logger = LoggerFactory.getLogger(getClass());

    /**
     * Digest of message contents for deterministic signing.
     */
    SHA256Digest msgDigest = new SHA256Digest();

    /**
     * The actual signing implementation. (used only signing -- superclass is used for verification)
     */
    ECDSASigner deterministicSigner;

    /**
     * Output the DER encoding of the ASN.1 sequence r,s
     * @param r
     * @param s
     * @return
     */
    public static byte[] derEncodeSignature(BigInteger r, BigInteger s) {
        ASN1Integer[] rs = {new ASN1Integer(r), new ASN1Integer(s)};
        DERSequence seq = new DERSequence(rs);

        try {
            return seq.getEncoded();
        } catch (IOException e) {
            throw new RuntimeException("Should never happen! DER Encoding exception", e);
        }
    }

    public ECDSADeterministicSignature() {
        DSAKCalculator kCalk = new HMacDSAKCalculator(new org.bouncycastle.crypto.digests.SHA256Digest());
        deterministicSigner = new ECDSASigner(kCalk);
    }

    @Override
    public void loadSigningCertificate(KeyStore.Builder keyStoreBuilder)
            throws CertificateException, UnrecoverableKeyException, IOException {
        super.loadSigningCertificate(keyStoreBuilder);

        if (!(loadedSigningKey instanceof ECPrivateKey)) {
            logger.error("Wrong private key type (expected ECPrivateKey, got {})", loadedSigningKey.getClass());
            throw new CertificateException("Wrong signing key type!");
        }
        ECPrivateKey key = (ECPrivateKey) loadedSigningKey;

        AsymmetricKeyParameter baseParams;
        try {
            baseParams = ECUtil.generatePrivateKeyParameter(key);
        } catch (InvalidKeyException e) {
            throw new UnrecoverableKeyException("Couldn't convert private key");
        }

        if (!(baseParams instanceof ECPrivateKeyParameters)) {
            logger.error("Error converting to bouncycastle type! (got {})", baseParams.getClass());
            throw new UnrecoverableKeyException("Wrong signing key type!");
        }

        ECPrivateKeyParameters params = (ECPrivateKeyParameters) baseParams;

        deterministicSigner.init(true, params);
    }

    /**
     * Add the list of messages to the stream that is being verified/signed.
     * Messages are prepended with their length in 32-bit big-endian format.
     *
     * @param msg
     * @throws SignatureException
     */
    @Override
    public void updateContent(Message msg) throws SignatureException {
        assert msg != null;

        // We're doing twice the digest work so that we also support verification with the same update.
        // If this becomes a problem, we can decide which way to update based on which init was called.
        super.updateContent(msg);
        msgDigest.update(msg);
    }

    @Override
    public void updateContent(InputStream in) throws IOException, SignatureException {
        ByteString inStr = ByteString.readFrom(in);
        signer.update(inStr.asReadOnlyByteBuffer());
        msgDigest.update(inStr);
    }

    @Override
    public Signature sign() throws SignatureException {
        Signature.Builder sig = Signature.newBuilder();
        sig.setType(Crypto.SignatureType.ECDSA);

        BigInteger[] rawSig = deterministicSigner.generateSignature(msgDigest.digest());

        sig.setData(ByteString.copyFrom(derEncodeSignature(rawSig[0], rawSig[1])));

        sig.setSignerId(loadedSigningKeyId);
        return sig.build();
    }
}
