package meerkat.crypto.concrete;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import com.google.protobuf.Message;
import meerkat.crypto.Encryption;
import meerkat.protobuf.ConcreteCrypto;
import meerkat.protobuf.Crypto;
import org.bouncycastle.crypto.params.AsymmetricKeyParameter;
import org.bouncycastle.crypto.params.ECDomainParameters;
import org.bouncycastle.crypto.params.ECKeyParameters;
import org.bouncycastle.crypto.params.ECPublicKeyParameters;
import org.bouncycastle.crypto.util.PublicKeyFactory;
import org.bouncycastle.jce.spec.ECParameterSpec;
import org.bouncycastle.math.ec.ECCurve;
import org.bouncycastle.math.ec.ECPoint;
import org.bouncycastle.util.BigIntegers;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import qilin.primitives.concrete.ECElGamal;
import qilin.primitives.concrete.ECGroup;
import qilin.util.PRGRandom;
import qilin.util.Pair;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.spec.*;
import java.util.Random;

/**
 * Created by talm on 17/11/15.
 */
public class ECElGamalEncryption extends GlobalCryptoSetup implements Encryption {
    final Logger logger = LoggerFactory.getLogger(getClass());

    public final static String KEY_ALGORITHM = "ECDH";

    /**
     * The Qilin format El-Gamal public key
     */
    ECElGamal.PK elGamalPK;

    ECCurve curve;

    ECGroup group;

    public ECGroup getGroup() { return group; }

    public ECElGamal.PK getElGamalPK() {
        return elGamalPK;
    }

    public void init(ConcreteCrypto.ElGamalPublicKey serializedPk) throws InvalidKeySpecException {
        AsymmetricKeyParameter keyParam;

        try {
            keyParam = PublicKeyFactory.createKey(serializedPk.getSubjectPublicKeyInfo().toByteArray());
        } catch (IOException e) {
            // Shouldn't every happen
            logger.error("Invalid Public Key Encoding", e);
            throw new InvalidKeySpecException("Invalid Public Key Encoding", e);
        }

        if (!(keyParam instanceof ECPublicKeyParameters)) {
            logger.error("Public key is a {}, not a valid public EC Key!", keyParam.getClass());
            throw new InvalidKeySpecException("Not a valid EC public key!");
        }

        ECDomainParameters params = ((ECKeyParameters) keyParam).getParameters();
        ECParameterSpec ecParams = new ECParameterSpec(params.getCurve(), params.getG(), params.getN(), params.getH(),
                params.getSeed());

        curve = params.getCurve();
        group = new ECGroup(ecParams);

        elGamalPK = new ECElGamal.PK(group, ((ECPublicKeyParameters) keyParam).getQ());
    }


    @Override
    public Crypto.RerandomizableEncryptedMessage encrypt(Message plaintext, Crypto.EncryptionRandomness rnd) {

        // We write the message using writeDelimited to so the length gets prepended.
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        try {
            plaintext.writeDelimitedTo(out);
        } catch (IOException e) {
            logger.error("Should never happen!", e);
            throw new RuntimeException("Error in ByteArrayOutputStream!", e);
        }
        byte[] msg = out.toByteArray();
        ECPoint encodedMsg = group.injectiveEncode(msg, new PRGRandom(msg));

        BigInteger rndInt = BigIntegers.fromUnsignedByteArray(rnd.getData().toByteArray());
        Pair<ECPoint,ECPoint> cipherText = elGamalPK.encrypt(encodedMsg, rndInt);
        ConcreteCrypto.ElGamalCiphertext encodedCipherText = ConcreteCrypto.ElGamalCiphertext.newBuilder()
                .setC1(ByteString.copyFrom(cipherText.a.getEncoded(true)))
                .setC2(ByteString.copyFrom(cipherText.b.getEncoded(true)))
                .build();

        return Crypto.RerandomizableEncryptedMessage.newBuilder()
                .setData(encodedCipherText.toByteString())
                .build();
    }

    @Override
    public Crypto.RerandomizableEncryptedMessage rerandomize(Crypto.RerandomizableEncryptedMessage msg, Crypto.EncryptionRandomness rnd) throws InvalidProtocolBufferException {
        BigInteger rndInt = BigIntegers.fromUnsignedByteArray(rnd.getData().toByteArray());
        Pair<ECPoint,ECPoint> randomizer = elGamalPK.encrypt(curve.getInfinity(), rndInt);
        ConcreteCrypto.ElGamalCiphertext originalEncodedCipher= ConcreteCrypto.ElGamalCiphertext.parseFrom(msg.getData());

        Pair<ECPoint,ECPoint> originalCipher = new Pair<>(
                curve.decodePoint(originalEncodedCipher.getC1().toByteArray()),
                curve.decodePoint(originalEncodedCipher.getC2().toByteArray()));
        Pair<ECPoint,ECPoint> newCipher = elGamalPK.add(originalCipher, randomizer);

        return Crypto.RerandomizableEncryptedMessage.newBuilder()
                .setData(
                    ConcreteCrypto.ElGamalCiphertext.newBuilder()
                    .setC1(ByteString.copyFrom(newCipher.a.getEncoded(true)))
                    .setC2(ByteString.copyFrom(newCipher.b.getEncoded(true)))
                    .build().toByteString()
                ).build();
    }

    @Override
    public Crypto.EncryptionRandomness generateRandomness(Random rand) {
        BigInteger randomInt = new BigInteger(group.getCurveParams().getN().bitLength() - 1, rand);
        Crypto.EncryptionRandomness retval = Crypto.EncryptionRandomness.newBuilder()
                .setData(ByteString.copyFrom(BigIntegers.asUnsignedByteArray(randomInt))).build();

        return retval;
    }
}
